/*****************************************************************************/
/*
                                  comment.c

Comments (reader feedback).


$$COMMENT$$
-----------
The following are the reserved character directives associated with
$$COMMENT$$.  Changes to these must be reflected in a couple of places in
COMMENT.C  Suggest searching for 0x21!

  '!'   further comment closed
  '^'   show comment(s) immediately
  '?'   show make-comment dialog immediately
  '_'   use fixed font to display comments
  ':'   make things that look like URLs into anchors
  '~'   example only (used by CommentSubmit())
  '%'   strftime() format string (used by CommentSubmit())
  '+'   comment is moderated (used by CommentSubmit())
  '='   comment is notified by email (used by CommentSubmit())
  digit is textarea rows-by-cols


COMMENT ABUSE
-------------
Apart from human moderation there are several mechanisms available for
configuration that attempt to minimise the possibility of comment abuse and/or
DOS attack.  These are (as-yet) unproven and relatively light-weight
approaches.

The user-agent is present and does not contain any of the line-separated
[comment-reject-agent] strings.

  http://www.botsvsbrowsers.com/

The comment 'who' cannot contain any of (seen in the wild):

  "http://"
  "https://"

[comment-host-max] sets a maximum number of comments allowed to originate from
the one host (IP address).

[comment-max-size] sets the maximum size (specified in kBytes) of of any one
comment file.  This includes all visible, cut and awaiting-moderation items. 
There is a hard-wired, absolute maximum file size applied.

[comment-max-chars] sets the maximum size (in bytes) of any one comment.

[comment-min-interval] specifies the minimum interval (in seconds) between
successive comments from the one host (IP address).  The current comment file
is scanned looking for all items from that host and is any are more recent the
post is rejected.

[comment-reject-string] allows a list of line-separated strings to be checked
for in comment text.  These strings can contain white-space and any other
printable character.  An example might be "[url=".  User feedback (email) also
uses these reject strings.

[comment-reject-word] allows a list of space-separated 'words' (really
character strings) to be checked for in the comment text.  If found the comment
is rejected.  CAUTION when constructing the 'word' list.  The comment text has
ALL white space, punctuation and digits removed before 'word' search.  This is
done to attempt to workaround obfuscation of abusive words.  Of course it
concatenates all significant characters which may produce unexpected results. 
User feedback (email) also uses these reject words.

If a comment page also has a $$TICKET$$ directive then the text must contain
the ticket string.  This attempts to impede SPAMbot submissions.  See
TicketFile() for implementation comment.


COMMENT FILE STRUCTURE
----------------------
A structure XML-like layout with element content HTML-escaped.

  <comment id="0123456789abcdef" moderation="*">
  <host>gort.private.net 192.168.1.2</host>
  <agent>blah-blah</agent>
  <when>Sun, 22 Feb 2009 06:45</when>
  <who>Mark.Daniel@wasd.vsm.com.au</who>
  <text>Comment text ...</text>
  </comment>

The moderation="." character indicates:

  '*'  unmoderated
  '+'  waiting for moderation
  '!'  moderation approved
  'X'  cut from comments (not deleted as such)
  'x'  un-cut back into comments


COPYRIGHT
---------
Copyright (C) 2008-2011 Mark G.Daniel
This program, comes with ABSOLUTELY NO WARRANTY.
This is free software, and you are welcome to redistribute it under the
conditions of the GNU GENERAL PUBLIC LICENSE, version 3, or any later version.


VERSION HISTORY
---------------
26-FEB-2011  MGD  configuration mail transport
09-DEC-2009  MGD  comment files now chmod(0700) (only owner access)
25-NOV-2009  MGD  add <agent>..</agent> (for documentation/SPAM purposes)
                  CommentSubmitVet() various additional constraints
20-APR-2009  MGD  getting to be bigger than Ben Hur
05-JAN-2007  MGD  initial
*/

/*****************************************************************************/

#ifdef VWCMS_VMS_V62
#undef _VMS_V6_SOURCE
#define _VMS_V6_SOURCE
#undef __VMS_VER
#define __VMS_VER 60200000
#undef __CRTL_VER
#define __CRTL_VER 60200000
#endif

#ifndef __VAX
#   pragma nomember_alignment
#endif

/* standard C header files */
#include <ctype.h>
#include <errno.h>
#include <stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

/* VMS related header files */
#include <libdef.h>
#include <libdtdef.h>
#include <lib$routines.h>
#include "rmsdef.h"
#include "ssdef.h"
#include "stsdef.h"

/* application header file */
#include "cgilib.h"
#include "comment.h"
#include "config.h"
#include "content.h"
#include "cookie.h"
#include "data.h"
#include "other.h"
#include "sendmail.h"
#include "template.h"
#include "ticket.h"
#include "vwcms.h"

#define FI_LI "COMMENT", __LINE__

/* global storage */

int  CommentTemplateCount;

/* prototypes */

/* external storage */

extern BOOL  Debug,
             ManageSiteOn,
             QueryComment,
             QueryPrint,
             QueryPrintPage,
             RequestMethodPOST,
             RequestSubmitOK,
             WatchEnabled;

extern unsigned long  CurrentSecondsTime;

extern unsigned long  CurrentVmsBinTime[];

extern char  *CgiRequestUriPtr,
             *CgiHttpUserAgentPtr;

extern char  ContentSectionName[],
             ContentSiteSlashName[],
             SoftwareId[],
             Utility[],
             LoginUserName[];

extern CONTENT_DATA  *ContentSectionPtr;

extern CONFIG_DATA  VWcmsConfig;

/*****************************************************************************/
/*
Successive calls display the next comment.  Returns true is a comment
displayed, false if exhausted.  First call returns true is there are comments
to be displayed, false is none.
*/

BOOL CommentGet
(
char *CommentName,
BOOL IncludeHost,
BOOL AnchorUrls,
BOOL FixedFont
)
{
   static char  *ContextPtr;

   BOOL  IsCut,
         IsUnModerated;
   int  status,
        TextLength;
   char  *cptr,
         *AgentPtr,
         *FileNamePtr,
         *HostPtr,
         *IdPtr,
         *ModerationPtr,
         *TextPtr,
         *WhenPtr,
         *WhoPtr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "CommentGet() |%s|\n", CommentName);

   if (CommentName)
   {
      /**************/
      /* first call */
      /**************/

      FileNamePtr = MakeFileName (CommentName, ".CMT");

      if (WatchEnabled)
         WatchThis (FI_LI, "COMMENT get !AZ !AZ", CommentName, FileNamePtr);

      status = WaitForLock (FileNamePtr, FALSE);
      if (VMSnok (status))
      {
         TemplateAdvise ("ECOMMENT !AZ (%X!8XL)", SysGetMsg(status), status);
         return (FALSE);
      }

      status = DataReadFileIntoMemory (FileNamePtr, &TextPtr,
                                       &TextLength, TRUE);

      WaitForUnLock ();

      if (VMSnok (status))
      {
         if (status == RMS$_FNF) return (FALSE);
         TemplateAdvise ("ECOMMENT !AZ (%X!8XL)", SysGetMsg(status), status);
         return (FALSE);
      }

      /* local copy because there'll be plenty of chopping going on */
      ContextPtr = CgiLibVeeMemCalloc (TextLength);
      if (!ContextPtr) ErrorExit (vaxc$errno, FI_LI);
      memcpy (ContextPtr, TextPtr, TextLength);

      /* quick-and-dirty guess for whether there are any comments */
      for (cptr = ContextPtr; *cptr; cptr++)
      {
         if (*cptr != '<') continue;
         if (strncmp (cptr, "<comment id=\"", 13)) continue;
         for (cptr += 13; *cptr && *cptr != '\"'; cptr++);
         if (!*cptr) break;
         for (cptr++; *cptr && isspace(*cptr); cptr++);
         if (strncmp (cptr, "moderation=\"", 12)) continue;
         cptr += 12;
         if (*cptr == 'X') continue;
         return (TRUE);
      }
      return (FALSE);
   }

   if (!(cptr = ContextPtr)) return (FALSE);

   while (*cptr)
   {
      /*****************/
      /* parse comment */
      /*****************/

      while (*cptr && *cptr != '<') cptr++;
      if (!*cptr) break;
      if (strncmp (cptr++, "<comment id=\"", 13)) continue;
      IdPtr = (cptr += 12);
      while (*cptr && *cptr != '\"') cptr++;
      if (cptr - IdPtr != 16) continue;
      *cptr = '\0';
      for (cptr++; *cptr && isspace(*cptr); cptr++);
      if (strncmp (cptr, "moderation=\"", 12)) continue;
      ModerationPtr = (cptr += 12);
      while (*cptr && *cptr != '\"') cptr++;
      if (cptr - ModerationPtr != 1) continue;
      while (*cptr && *cptr != '>') cptr++;
      if (!*cptr) break;

      /* if the comment has been cut (deleted) */
      IsCut = (*ModerationPtr == 'X');
      if (!ManageSiteOn && IsCut) continue;

      /* note if it's unmoderated */
      IsUnModerated = (*ModerationPtr == '+');

      while (*cptr && *cptr != '<') cptr++;
      if (strncmp (cptr, "<host>", 6)) continue;
      HostPtr = (cptr += 6);
      while (*cptr && *cptr != '<') cptr++;
      if (strncmp (cptr, "</host>", 7)) continue;
      *cptr = '\0';
      cptr += 7;

      while (*cptr && *cptr != '<') cptr++;
      if (!strncmp (cptr, "<agent>", 7))
      {
         /* agent was not present in earlier versions of the comment entity */
         AgentPtr = (cptr += 7);
         while (*cptr && *cptr != '<') cptr++;
         if (strncmp (cptr, "</agent>", 8)) continue;
         *cptr = '\0';
         cptr += 8;
      }

      while (*cptr && *cptr != '<') cptr++;
      if (strncmp (cptr, "<when>", 6)) continue;
      WhenPtr = (cptr += 6);
      while (*cptr && *cptr != '<') cptr++;
      if (strncmp (cptr, "</when>", 7)) continue;
      *cptr = '\0';
      cptr += 7;

      while (*cptr && *cptr != '<') cptr++;
      if (strncmp (cptr, "<who>", 5)) continue;
      WhoPtr = (cptr += 5);
      while (*cptr && *cptr != '<') cptr++;
      if (strncmp (cptr, "</who>", 6)) continue;
      *cptr = '\0';
      cptr += 6;

      while (*cptr && *cptr != '<') cptr++;
      if (strncmp (cptr, "<email>", 7)) continue;
      cptr += 7;
      /* not for publication */
      while (*cptr && *cptr != '<') cptr++;
      if (strncmp (cptr, "</email>", 8)) continue;
      *cptr = '\0';
      cptr += 8;

      while (*cptr && *cptr != '<') cptr++;
      if (strncmp (cptr, "<text>", 6)) continue;
      TextPtr = (cptr += 6);
      while (*cptr && *cptr != '<') cptr++;
      if (strncmp (cptr, "</text>", 7)) continue;
      *cptr = '\0';
      cptr += 7;

      while (*cptr && *cptr != '<') cptr++;
      if (strncmp (cptr, "</comment>", 10)) continue;

      /* save it for the next call */
      ContextPtr = cptr;

      /*******************/
      /* display comment */
      /*******************/

      /* munge (as-yet) unmoderated text (except for logged-in managers) */
      if (IsUnModerated && !ManageSiteOn)
         for (cptr = TextPtr; *cptr; cptr++)
            if (IS8PRINT(*cptr)) *cptr = '*';

      fputs ("<tr><td class=\"vwcms_comment_item\">\n", stdout);

      if (IsUnModerated)
         fprintf (stdout,
"<span class=\"vwcms_comment_unmoderated\">%s</span>\n",
                  CONFIG_LANG_PARSE(LangModerate,1));

      if (IncludeHost)
      {
         /* just the recorded host name */
         for (cptr = HostPtr; *cptr && *cptr != ' '; cptr++);
         *cptr = '\0';
         fprintf (stdout, "<span class=\"vwcms_comment_who\">%s @ %s</span>\n",
                  HtmlEscape(WhoPtr),
                  HtmlEscape(HostPtr));
      }
      else
         fprintf (stdout, "<span class=\"vwcms_comment_who\">%s</span>\n",
                  HtmlEscape(WhoPtr));

      if (AnchorUrls)
      {
         /* bring it back to plain-text */
         ContentDeEntify (TextPtr, FALSE);
         /* calloc()s memory so need to explicitly free it in case CGIplus */
         TextPtr = (char*)CgiLibAnchorHtmlEscape (TextPtr, -1, NULL, 0, TRUE);
      }

      fprintf (stdout,
"<span class=\"vwcms_comment_time\">%s</span>\n\
<span class=\"vwcms_comment_text\"%s%s%s%s>%s</span>\n",
               WhenPtr,
               (FixedFont || IsCut) ? " style=\"" : "",
               FixedFont ? FONT_FAMILY_MONOSPACE : "",
               IsCut ? "text-decoration:line-through;" : "",
               (FixedFont || IsCut) ? "\"" : "",
               TextPtr);

      if (AnchorUrls) free (TextPtr);

      if (ManageSiteOn)
      {
         if (IsCut)
            fprintf (stdout,
"&nbsp;<input type=\"checkbox\" name=\"mod_uncut_%s\" value=\"1\">&nbsp;%s\n",
                     IdPtr, CONFIG_LANG_PARSE(LangModerate,6));
         else
            fprintf (stdout,
"&nbsp;<input type=\"checkbox\" name=\"mod_cut_%s\" value=\"1\">&nbsp;%s\n",
                     IdPtr, CONFIG_LANG_PARSE(LangModerate,5));

         if (IsUnModerated)
            fprintf (stdout,
"&nbsp;<input type=\"checkbox\" name=\"mod_tick_%s\" value=\"1\">&nbsp;\
<span style=\"font-weight:bold;text-decoration:underline;\">%s</span>\n",
                     IdPtr, CONFIG_LANG_PARSE(LangModerate,4));
      }

      fputs ("</td></tr>\n", stdout);

      return (TRUE);
   }

   /*************/
   /* exhausted */
   /*************/

   ContextPtr = NULL;

   return (FALSE);
}

/*****************************************************************************/
/*
Display any comments (reader feedback) and prompt for user comment.
Called by a $$COMMENT$..$$ directive in TemplateLoad() or TemplateContent(). 
*/

void CommentTemplate (char* SubsArg)

{
   BOOL  AnchorUrls = FALSE,
         CommentClosed = FALSE,
         FixedFont = FALSE,
         IncludeHost = FALSE,
         IsComment,
         MakeCommentOpen = FALSE,
         OpenComments = FALSE;
   int  TextRows = 0,
        TextCols = 0;
   char  *cptr, *sptr, *zptr,
         *CommentPtr,
         *CommentNamePtr,
         *TestTicketPtr = NULL,
         *FromPtr,
         *FremailPtr,
         *HidePtr,
         *HideImagePtr,
         *MakePtr,
         *MakeImagePtr,
         *ModeratePtr,
         *ModerateImagePtr,
         *NotForPubPtr,
         *HostNamePtr,
         *SectionNamePtr,
         *ShowPtr,
         *ShowImagePtr,
         *SubmitPtr,
         *TextPtr,
         *WhoPtr,
         *WhoEmailPtr;
   char  CommentFremail [256] = "",
         CommentFrom [256] = "",
         CommentName [256] = "",
         HideComment [256] = "",
         MakeComment [256] = "",
         NotForPub [256] = "",
         ShowComment [256] = "",
         SubmitComment [256] = "";

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "CommentTemplate() |%s|\n", SubsArg);

   if (ContentSectionPtr)
      SectionNamePtr = ContentSectionPtr->SectionName;
   else
      SectionNamePtr = ContentSectionName;

   OpenComments = QueryComment;

   if (*(cptr = SubsArg) == '$') cptr++;

   while (*cptr)
   {
      while (*cptr && *cptr != '%' && !isalnum(*cptr))
      {
         if (*cptr == '!')
            CommentClosed = TRUE;
         else
         if (*cptr == '^')
            OpenComments = TRUE;
         else
         if (*cptr == '?')
            MakeCommentOpen = TRUE;
         else
         if (*cptr == '_')
            FixedFont = TRUE;
         else
         if (*cptr == '.')
            IncludeHost = TRUE;
         else
         if (*cptr == ':')
            AnchorUrls = TRUE;
         else
         if (*cptr == '~')
            TestTicketPtr = TicketTest (FALSE);
         cptr++;
      }
      if (!*cptr) break;

      if (*cptr == '%')
         cptr = TemplateParseSkipParam (cptr, NULL);
      else
      if (isdigit(*cptr))
         cptr = TemplateParseRowCol (cptr, &TextRows, &TextCols);
      else
      if (strsame (cptr, "email=", 6))
         /* only of interest to CommentSubmit() */
         cptr = TemplateParseSkipParam (cptr+6, NULL);
      else
      if (strsame (cptr, "fremail=", 8))
         cptr = TemplateParseParam (cptr+8, CommentFremail, sizeof(CommentFremail));
      else
      if (strsame (cptr, "from=", 5))
         cptr = TemplateParseParam (cptr+5, CommentFrom, sizeof(CommentFrom));
      else
      if (strsame (cptr, "hide=", 5))
         cptr = TemplateParseParam (cptr+5, HideComment, sizeof(HideComment));
      else
      if (strsame (cptr, "make=", 5))
         cptr = TemplateParseParam (cptr+5, MakeComment, sizeof(MakeComment));
      else
      if (strsame (cptr, "name=", 5))
         cptr = TemplateParseParam (cptr+5, CommentName, sizeof(CommentName));
      else
      if (strsame (cptr, "nfp=", 4))
         cptr = TemplateParseParam (cptr+4, NotForPub, sizeof(NotForPub));
      else
      if (strsame (cptr, "origin=", 7))
         cptr = TemplateParseSkipParam (cptr+7, NULL);
      else
      if (strsame (cptr, "show=", 5))
         cptr = TemplateParseParam (cptr+5, ShowComment, sizeof(ShowComment));
      else
      if (strsame (cptr, "size=", 5))
         cptr = TemplateParseRowCol (cptr+5, &TextRows, &TextCols);
      else
      if (strsame (cptr, "submit=", 7))
         cptr = TemplateParseParam (cptr+7, SubmitComment, sizeof(SubmitComment));
      else
      {
         TemplateParamError ("COMMENT", SubsArg, cptr);
         return;
      }
   }

   if (TextRows <= 0 || TextRows >= 24) TextRows = 16;
   if (TextCols <= 0 || TextCols >= 132) TextCols = 80;

   if (!*(CommentNamePtr = CommentName)) CommentNamePtr = SectionNamePtr;
   if (!*(ShowPtr = ShowComment)) ShowPtr = CONFIG_LANG_PARSE(LangComment,1);
   if (!*(HidePtr = HideComment)) HidePtr = CONFIG_LANG_PARSE(LangComment,2);
   if (!*(MakePtr = MakeComment)) MakePtr = CONFIG_LANG_PARSE(LangComment,3);
   if (!*(SubmitPtr = SubmitComment))
      SubmitPtr = CONFIG_LANG_PARSE(LangComment,4);
   if (!*(FromPtr = CommentFrom)) FromPtr = CONFIG_LANG_PARSE(LangComment,5);
   if (!*(FremailPtr = CommentFremail))
      FremailPtr = CONFIG_LANG_PARSE(LangComment,6);
   if (!*(NotForPubPtr = NotForPub))
      NotForPubPtr = CONFIG_LANG_PARSE(LangComment,7);

   /*******/
   /* who */
   /*******/

   if (IncludeHost)
   {
      cptr = HostNamePtr = GetCgiVar ("REMOTE_HOST");
      while (*cptr && (*cptr == '.' || isdigit(*cptr))) cptr++;
      if (!*cptr) HostNamePtr = HostNameByAddr (HostNamePtr);
   }
   else
      HostNamePtr = NULL;

   WhoPtr = CookieGet (COMMENT_WHO_COOKIE);
   if (WhoPtr)
   {
      CgiLibUrlDecode (WhoPtr);
      for (cptr = sptr = WhoPtr; *cptr; cptr++)
         if (IS8PRINT(*cptr)) *sptr++ = *cptr;
      *sptr = '\0';
      if (!*WhoPtr) WhoPtr = NULL;
   }
   if (!WhoPtr) WhoPtr = FromPtr;

   WhoEmailPtr = CookieGet (COMMENT_EMAIL_COOKIE);
   if (WhoEmailPtr)
   {
      CgiLibUrlDecode (WhoEmailPtr);
      for (cptr = sptr = WhoEmailPtr; *cptr; cptr++)
         if (IS8PRINT(*cptr)) *sptr++ = *cptr;
      *sptr = '\0';
      if (!*WhoEmailPtr) WhoEmailPtr = NULL;
   }
   if (!WhoEmailPtr) WhoEmailPtr = FremailPtr;

   if (!CommentTemplateCount++)
   {
      /************************/
      /* JavaScript only once */
      /************************/

      fputs ("<script type=\"text/javascript\">\n\
function toggleShowComment (num) {\n\
   var odiv = document.getElementById(\"comment_\"+num+\"_show_div\");\n\
   var osbtn = document.getElementById(\"comment_\"+num+\"_show_btn\");\n\
   var ohbtn = document.getElementById(\"comment_\"+num+\"_hide_btn\");\n\
   if (odiv.style.display == \"none\") {\n\
      odiv.style.display = \"block\";\n\
      osbtn.style.display = \"none\";\n\
      ohbtn.style.display = \"inline\";\n\
   }\n\
   else {\n\
      odiv.style.display = \"none\";\n\
      osbtn.style.display = \"inline\";\n\
      ohbtn.style.display = \"none\";\n\
   }\n\
}\n\
function toggleMakeComment (num) {\n\
   var odiv = document.getElementById(\"comment_\"+num+\"_make_div\");\n\
   if (odiv.style.display == \"none\") {\n\
      odiv.style.display = \"block\";\n\
      window.scrollTo (0, document.getElementById(\
\"comment_\"+num+\"_make_btn\").offsetTop);\n\
   }\n\
   else\n\
      odiv.style.display = \"none\";\n\
}\n",
             stdout);

   if (ManageSiteOn)
      fprintf (stdout,
"function areYouSureModerate () {\n\
   return confirm(\"%s.  %s\");\n\
}\n",
               CONFIG_LANG_PARSE(LangModerate,3),
               CONFIG_LANG_PARSE(LangManage6,1));

      fputs ("</script>\n", stdout);
   }

   /***********/
   /* comment */
   /***********/

   IsComment = CommentGet (CommentNamePtr, FALSE, FALSE, FALSE);

   /* only display comments when printing a single page */
   if (QueryPrintPage)
      OpenComments = TRUE;
   else
   if (QueryPrint)
      OpenComments = FALSE;

   fprintf (stdout,
"<div style=\"width:100%%\">\n\
<form class=\"vwcms_comment\" style=\"display:none\" \
id=\"comment_%d_form\" name=\"comment_form\" \
method=\"POST\" action=\"%s/%s?comment\">\n\
<input type=\"hidden\" name=\"comment_name\" value=\"%s\">\n\
<input type=\"hidden\" name=\"comment_cols\" value=\"%d\">\n",
            CommentTemplateCount,
            UrlEncodeName(ContentSiteSlashName),
            UrlEncodeName(SectionNamePtr),
            HtmlEscape(CommentNamePtr),
            TextCols);

   if (TestTicketPtr)
      fprintf (stdout,
"<input type=\"hidden\" name=\"example_ticket\" value=\"%s\">\n",
               TestTicketPtr); 

   /************/
   /* show btn */
   /************/

   if (ShowImagePtr = IsAnImage(ShowPtr))
   {
      if (!(HideImagePtr = IsAnImage(HidePtr))) HideImagePtr = HidePtr;

      fprintf (stdout,
"<input id=\"comment_%d_show_btn\" \
class=\"vwcms_btn\" onclick=\"toggleShowComment(%d);return false;\" \
type=\"image\" %s style=\"margin-right:1em;border-style:none;\"%s />\n\
\
<input id=\"comment_%d_hide_btn\" \
class=\"vwcms_btn\" onclick=\"toggleShowComment(%d);return false;\" \
type=\"image\" %s style=\"margin-right:1em;border-style:none;\"%s />\n",
               CommentTemplateCount,
               CommentTemplateCount,
               ShowImagePtr,
               IsComment ? "" : " disabled",
               CommentTemplateCount,
               CommentTemplateCount,
               HideImagePtr,
               IsComment ? "" : " disabled");
   }
   else
   {
      fprintf (stdout,
"<input id=\"comment_%d_show_btn\" \
class=\"vwcms_btn\" style=\"margin-right:1em;\" \
onclick=\"toggleShowComment(%d);return false;\" \
type=\"submit\" value=\"%s\"%s />\n\
\
<input id=\"comment_%d_hide_btn\" \
class=\"vwcms_btn\" style=\"margin-right:1em;\" \
onclick=\"toggleShowComment(%d);return false;\" \
type=\"submit\" value=\"%s\"%s />\n",
               CommentTemplateCount,
               CommentTemplateCount,
               ShowPtr,
               IsComment ? "" : " disabled",
               CommentTemplateCount,
               CommentTemplateCount,
               HidePtr,
               IsComment ? "" : " disabled");
   }

   /*********************/
   /* actual comment(s) */
   /*********************/

   fprintf (stdout,
"<div id=\"comment_%d_show_div\" style=\"display:none;\">\n\
<table class=\"vwcms_comment_list\" \
style=\"margin-top:1em;margin-bottom:1em;\" \
border=\"0\" cellpadding=\"0\" cellspacing=\"0\">\n",
            CommentTemplateCount);

   while (CommentGet (NULL, IncludeHost, AnchorUrls, FixedFont));

   fputs ("</table>\n", stdout);

   if (ManageSiteOn)
   {
      /****************/
      /* moderate btn */
      /****************/

      ModeratePtr = CONFIG_LANG_PARSE(LangModerate,3);

      if (ModerateImagePtr = IsAnImage(ModeratePtr))
         fprintf (stdout,
"<input id=\"comment_%d_moderate_btn\" \
class=\"vwcms_btn\" style=\"margin-bottom:1em;\" \
onclick=\"return areYouSureModerate();\" \
type=\"image\" name=\"comment_moderate\" %s style=\"border-style:none;\" />\n",
                  CommentTemplateCount,
                  ModerateImagePtr);
      else
         fprintf (stdout,
"<input id=\"comment_%d_moderate_btn\" \
class=\"vwcms_btn\" style=\"margin-bottom:1em;\" \
onclick=\"return areYouSureModerate();\" \
type=\"submit\" name=\"comment_moderate\" value=\"%s\" />\n",
                  CommentTemplateCount,
                  ModeratePtr);
   }

   fputs ("</div>\n", stdout);

   /***************/
   /* make dialog */
   /***************/

   if (RequestMethodPOST)
   {
      /* should only happen with examples */
      MakeCommentOpen = TRUE;
      TextPtr = GetCgiVarNull ("FORM_COMMENT_TEXT");
      if (sptr = GetCgiVarNull ("FORM_EXAMPLE_TICKET"))
         if (cptr = strstr (TextPtr, sptr))
            TicketExcise (TextPtr, cptr);
   }
   else
      TextPtr = "";

   if (MakeImagePtr = IsAnImage(MakePtr))
      fprintf (stdout,
"<input id=\"comment_%d_make_btn\" \
class=\"vwcms_btn\" onclick=\"toggleMakeComment(%d);return false;\" \
type=\"image\" %s style=\"border-style:none;\"%s />\n",
               CommentTemplateCount,
               CommentTemplateCount,
               MakeImagePtr,
               CommentClosed ? " disabled" : "");
   else
      fprintf (stdout,
"<input id=\"comment_%d_make_btn\" \
class=\"vwcms_btn\" onclick=\"toggleMakeComment(%d);return false;\" \
type=\"submit\" value=\"%s\"%s />\n",
               CommentTemplateCount,
               CommentTemplateCount,
               MakePtr,
               CommentClosed ? " disabled" : "");

   fprintf (stdout,
"<div id=\"comment_%d_make_div\" style=\"display:none;\">\n\
<input style=\"margin-top:1em;\" name=\"comment_who\" \
onfocus=\"this.select()\" \
type=\"text\" size=\"%d\" value=\"%s\" /><b>%s%s</b><br />\n\
<input style=\"margin-top:1em;\" name=\"comment_email\" \
onfocus=\"this.select()\" \
type=\"text\" size=\"%d\" value=\"%s\" /> %s\n\
<textarea name=\"comment_text\" style=\"margin-top:1em;white-space:normal;\" \
rows=\"%d\" cols=\"%d\">%s</textarea><br />\n\
<input class=\"vwcms_btn\" style=\"margin-top:1em;\" name=\"comment_make\" ",
            CommentTemplateCount,
            TextCols / 2,
            HtmlEscape(WhoPtr),
            HostNamePtr ? "&nbsp;@&nbsp;" : "",
            HostNamePtr ? HostNamePtr : "",
            TextCols / 2,
            HtmlEscape(WhoEmailPtr),
            NotForPubPtr,
            TextRows, TextCols,
            HtmlEscape(TextPtr));

   if (cptr = IsAnImage(SubmitPtr))
      fprintf (stdout, "type=\"image\" %s />\n", cptr);
   else
      fprintf (stdout, "type=\"submit\" value=\"%s\" />\n", SubmitPtr);

   /* with comments always advise close to the submit button */
   TemplateAdvise ("$$$$<span style=\"font-size:80%\">$</span>");

   /**************/
   /* javascript */
   /**************/

   /* this just keeps it's hidden from non-JavaScript browsers */
   fprintf (stdout,
"<script type=\"text/javascript\">\
document.getElementById(\"comment_%d_form\").style.display = \"inline\";\
document.getElementById(\"comment_%d_hide_btn\").style.display = \"none\";",
            CommentTemplateCount,
            CommentTemplateCount);

   if (OpenComments && IsComment)
   {
      fprintf (stdout, "toggleShowComment(%d);", CommentTemplateCount);

      if (QueryComment)
         fprintf (stdout, "window.scrollTo(0,document.getElementById(\
\"comment_%d_make_btn\").offsetTop);\n",
                  CommentTemplateCount);
   }

   if (MakeCommentOpen)
      fprintf (stdout, "toggleMakeComment(%d);", CommentTemplateCount);

   if (QueryPrintPage)
      fprintf (stdout,
"document.getElementById(\"comment_%d_show_btn\").style.display = \"none\";\
document.getElementById(\"comment_%d_make_btn\").style.display = \"none\";",
               CommentTemplateCount,
               CommentTemplateCount);

   fputs ("</script>\n</div>\n</form>\n", stdout);

   if (ShowImagePtr)
      fprintf (stdout,
"<noscript>\
<form>\
<input class=\"vwcms_btn\" \
type=\"image\" %s style=\"margin-right:1em;border-style:none;\" disabled />\
<input class=\"vwcms_btn\" \
type=\"image\" %s style=\"margin-right:1em;border-style:none;\" disabled />\
<b>[JavaScript!]</b>\
</form>\
</noscript>\n",
               ShowImagePtr,
               MakeImagePtr);
   else
      fprintf (stdout,
"</script>\n\
<noscript>\
<form>\
<input class=\"vwcms_btn\" style=\"margin-right:1em;\" \
type=\"reset\" value=\"%s\" disabled />\
<input class=\"vwcms_btn\" style=\"margin-right:1em;\" \
type=\"reset\" value=\"%s\" disabled /><b>[JavaScript!]</b>\
</form>\
</noscript>\n",
               ShowPtr,
               MakePtr);

   fputs ("</div>\n", stdout);
}

/*****************************************************************************/
/*
Append the submitted comment to the section's comment list file.
*/

BOOL CommentSubmit ()

{
   BOOL  CommentModerated = FALSE;
   int  status,
        MailTextLength,
        ExampleOnly = FALSE,
        TextCols = 0,
        TextLength = 0;
   char  *cptr, *sptr, *tptr, *zptr,
         *CommentNamePtr,
         *TestTicketPtr = NULL,
         *FileNamePtr,
         *HostNamePtr,
         *MailTextPtr,
         *RemoteHostPtr,
         *TextPtr,
         *TimeStringPtr,
         *WhoPtr,
         *WhoEmailPtr;
   char  CommentEmail [256] = "",
         MailFrom [256],
         SiteOrigin [256] = "",
         SubjText [256],
         TimeFormat [256] = "";
   FILE  *FilePtr;
   DYSTRING  CommentDirect;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "CommentSubmit()\n");

   if (ConfigReadOnly (TRUE)) return (FALSE);

   CommentNamePtr = GetCgiVarNull ("FORM_COMMENT_NAME");
   if (!CommentNamePtr) ErrorExit (SS$_BUGCHECK, FI_LI);

   /* ensure there is such a comment */ 
   if (!(tptr = CommentSubmitFind (CommentNamePtr)))
   {
      if (WatchEnabled) WatchThis (FI_LI, "REJECT not found");
      TemplateAdvise ("W!AZ ($$COMMENT$!AZ$$\?\?\?)",
                      CONFIG_LANG_PARSE(LangModerate,2), CommentNamePtr);
      return (FALSE);
   }

   /* using request-persistent storage */
   DYSTRING_VEEMEM_INIT(&CommentDirect,0);

   TemplateParseDirect (&CommentDirect, tptr);

   cptr = DYSTRING_PTR(&CommentDirect);
   if (WatchEnabled) WatchThis (FI_LI, "COMMENT !AZ", cptr);

   while (*cptr)
   {
      while (*cptr && *cptr != '%' && !isalnum(*cptr))
      {
         if (*cptr == '!')
         {
            /* comments are (perhaps now) closed */
            TemplateAdvise ("W!AZ ($$COMMENT$!!$$)",
                            CONFIG_LANG_PARSE(LangModerate,2));
            return (FALSE);
         }
         if (*cptr == '~')
         {
            ExampleOnly = TRUE;
            TestTicketPtr = GetCgiVarNull ("FORM_EXAMPLE_TICKET");
         }
         else
         if (*cptr == '+')
            CommentModerated = TRUE;
         if (*cptr != '$') cptr++;
         if (*cptr == '$') cptr++;
      }

      if (*cptr == '%')
         cptr = TemplateParseParam (cptr, TimeFormat, sizeof(TimeFormat));
      else
      if (isdigit(*cptr))
         cptr = TemplateParseRowCol (cptr, NULL, &TextCols);
      else
      if (strsame (cptr, "email=", 6))
         cptr = TemplateParseParam (cptr+6, CommentEmail, sizeof(CommentEmail));
      else
      if (strsame (cptr, "origin=", 7))
         cptr = TemplateParseParam (cptr+7, SiteOrigin, sizeof(SiteOrigin));
      else
      if (strsame (cptr, "size=", 5))
         cptr = TemplateParseRowCol (cptr+5, NULL, &TextCols);
      else
         /* just ignore any others used by TemplateComment() */
         cptr = TemplateParseSkipParam (cptr, NULL);
   }

   if (TextCols <= 0 || TextCols >= 132) TextCols = 80;

   /****************/
   /* comment data */
   /****************/

   TimeStringPtr = TimeString (TimeFormat, CurrentSecondsTime);

   cptr = GetCgiVarNull ("FORM_COMMENT_COLS");
   if (!cptr) ErrorExit (SS$_BUGCHECK, FI_LI);
   TextCols = atoi(cptr);
   if (TextCols < 32) TextCols = 32;
   if (TextCols > 255) TextCols = 255;

   WhoPtr = GetCgiVarNull ("FORM_COMMENT_WHO");
   if (!WhoPtr) ErrorExit (SS$_BUGCHECK, FI_LI);
   /* ensure it's reasonable */
   for (cptr = sptr = WhoPtr; *cptr; cptr++)
      if (IS8PRINT(*cptr)) *sptr++ = *cptr;
   if (sptr > WhoPtr + 48) sptr = WhoPtr + 48;
   *sptr = '\0';
   if (!*WhoPtr) WhoPtr = CONFIG_LANG_PARSE(LangComment,5);
   /* if they are different put it into a cookie */
   if (strcmp (WhoPtr, CONFIG_LANG_PARSE(LangComment,5)))
      CookieSet (COMMENT_WHO_COOKIE, UrlEncode(WhoPtr));

   WhoEmailPtr = GetCgiVarNull ("FORM_COMMENT_EMAIL");
   if (!WhoEmailPtr) ErrorExit (SS$_BUGCHECK, FI_LI);
   /* ensure it's reasonable */
   for (cptr = sptr = WhoEmailPtr; *cptr; cptr++)
      if (IS8PRINT(*cptr)) *sptr++ = *cptr;
   if (sptr > WhoEmailPtr + 48) sptr = WhoEmailPtr + 48;
   *sptr = '\0';
   if (!*WhoEmailPtr) WhoEmailPtr = CONFIG_LANG_PARSE(LangComment,6);
   /* if they are different put it into a cookie */
   if (strcmp (WhoEmailPtr, CONFIG_LANG_PARSE(LangComment,6)))
      CookieSet (COMMENT_EMAIL_COOKIE, UrlEncode(WhoEmailPtr));

   HostNamePtr = RemoteHostPtr = GetCgiVar ("REMOTE_HOST");
   if (!HostNamePtr) ErrorExit (SS$_BUGCHECK, FI_LI);
   for (cptr = HostNamePtr;
        *cptr && (*cptr == '.' || isdigit(*cptr));
        cptr++);
   if (!*cptr) cptr = HostNameByAddr (HostNamePtr);
   if (cptr) HostNamePtr = cptr; 

   TextPtr = GetCgiVarNull ("FORM_COMMENT_TEXT");
   if (!TextPtr) ErrorExit (SS$_BUGCHECK, FI_LI);
   TextPtr = TidyTextArea (TextPtr, TextCols);

   if (ExampleOnly)
   {
      /****************/
      /* example/test */
      /****************/

      if (CommentRejectWho (WhoPtr, CONFIG_LANG_PARSE(LangModerate,2)))
         return (FALSE);

      if (CommentRejectString (TextPtr, CONFIG_LANG_PARSE(LangModerate,2)))
         return (FALSE);

      if (CommentRejectWord (TextPtr, CONFIG_LANG_PARSE(LangModerate,2)))
         return (FALSE);

      if (TestTicketPtr)
      {
         if (cptr = strstr (TextPtr, TestTicketPtr))
            TicketExcise (TextPtr, cptr);
         else
         {
            TemplateAdvise ("W!AZ", CONFIG_LANG_PARSE(LangModerate,2));
            return (FALSE);
         }
      }

      RequestSubmitOK = TRUE;

      /* false in the sense we want to go back to the same page */
      TemplateAdvise ("I!AZ", CONFIG_LANG_PARSE(LangFeedback,4));
      return (FALSE);
   }

   /***************/
   /* vet comment */
   /***************/

   if (TicketPresent ())
   {
      /* there is a $$TICKET$$ on the page */
      status = TicketFile (NULL, TextPtr);
      if (VMSnok (status))
      {
         /* just add a little delay into the mix */
         sleep (REJECT_SLEEP);
         return (FALSE);
      }
   }

   FileNamePtr = MakeFileName (CommentNamePtr, ".CMT");

   if (!CommentSubmitVet (FileNamePtr, RemoteHostPtr, WhoPtr, TextPtr))
   {
      /* just add a little delay into the mix */
      sleep (REJECT_SLEEP);
      return (FALSE);
   }

   /******************/
   /* append comment */
   /******************/

   status = WaitForLock (FileNamePtr, TRUE);
   if (VMSnok (status))
   {
      TemplateAdvise ("ECOMMENT !AZ (%X!8XL)", SysGetMsg(status), status);
      return (FALSE);
   }

   FilePtr = fopen (FileNamePtr, "a", "ctx=rec",
                    "rfm=stmlf", "shr=nil", "fop=cif");
   if (!FilePtr)
   {
      status = vaxc$errno;
      WaitForUnLock ();
      TemplateAdvise ("ECOMMENT !AZ (%X!8XL)", SysGetMsg(status), status);
      return (FALSE);
   }

   fprintf (FilePtr,
"<comment id=\"%08.08x%08.08x\" moderation=\"%c\">\n\
<host>%s %s</host>\n\
<agent>%s</agent>\n\
<when>%s</when>\n\
<who>%s</who>\n\
<email>%s</email>\n\
<text>%s</text>\n\
</comment>\n\
\n",
             CurrentVmsBinTime[1], CurrentVmsBinTime[0],
             CommentModerated ? '+' : '*',
             HostNamePtr, RemoteHostPtr,
             HtmlEscape(CgiHttpUserAgentPtr),
             TimeStringPtr,
             HtmlEscape(WhoPtr),
             HtmlEscape(WhoEmailPtr),
             HtmlEscape(TextPtr));

   fclose (FilePtr);

   if (chmod (FileNamePtr, 0700)) ErrorExit (vaxc$errno, FI_LI);

   WaitForUnLock ();

   if (CommentEmail[0])
   {
      /********************/
      /* notify moderator */
      /********************/

      zptr = (sptr = MailFrom) + sizeof(MailFrom)-1;
      if (VWcmsConfig.PMDFkludge)
      {
         /* for example %IN"", up to and including the first '"' */
         for (cptr = VWcmsConfig.PMDFkludge;
              *cptr && sptr < zptr;
              *sptr++ = *cptr++)
            if (cptr > VWcmsConfig.PMDFkludge && *(cptr-1) == '\"') break;
      }
      if (*(cptr = SiteOrigin))
         while (*cptr && sptr < zptr) *sptr++ = *cptr++;
      else
      if (cptr = VWcmsConfig.SiteOrigin)
         while (*cptr && sptr < zptr) *sptr++ = *cptr++;
      else
      {
         for (cptr = "VWcms@"; *cptr && sptr < zptr; *sptr++ = *cptr++);
         cptr = GetCgiVarNull ("SERVER_NAME");
         if (!cptr || !*cptr) ErrorExit (SS$_BUGCHECK, FI_LI);
         while (*cptr && sptr < zptr) *sptr++ = *cptr++;
      }
      if (VWcmsConfig.PMDFkludge)
      {
         /* for example %IN"", everything following the first '"' */
         for (cptr = VWcmsConfig.PMDFkludge; *cptr; cptr++)
            if (cptr > VWcmsConfig.PMDFkludge && *(cptr-1) == '\"') break;
          while (*cptr && sptr < zptr) *sptr++ = *cptr++;
      }
      if (sptr >= zptr) ErrorExit (SS$_BUGCHECK, FI_LI);
      *sptr = '\0';

      zptr = (sptr = SubjText) + sizeof(SubjText)-1;
      for (cptr = "VWcms "; *cptr && sptr < zptr; *sptr++ = *cptr++);
      if (CommentModerated)
      {
         for (cptr = CONFIG_LANG_PARSE(LangModerate,3);
               *cptr && sptr < zptr;
               *sptr++ = *cptr++);
         if (sptr < zptr) *sptr++ = ' ';
      }
      for (cptr = CgiRequestUriPtr;
           *cptr && *cptr != '?' && sptr < zptr;
           *sptr++ = *cptr++);
      if (sptr >= zptr) ErrorExit (SS$_BUGCHECK, FI_LI);
      *sptr = '\0';

      MailTextLength = strlen(TextPtr) + 1024;
      MailTextPtr = sptr = CgiLibVeeMemCalloc (MailTextLength);
      if (!MailTextPtr) ErrorExit (vaxc$errno, FI_LI);
      zptr = sptr + MailTextLength;
      for (cptr = "http://"; *cptr && sptr < zptr; *sptr++ = *cptr++);
      cptr = GetCgiVarNull ("SERVER_NAME");
      if (!cptr) ErrorExit (SS$_BUGCHECK, FI_LI);
      while (*cptr && sptr < zptr) *sptr++ = *cptr++;
      for (cptr = CgiRequestUriPtr;
           *cptr && *cptr != '?' && sptr < zptr;
           *sptr++ = *cptr++);
      for (cptr = "\n\n"; *cptr && sptr < zptr; *sptr++ = *cptr++);
      for (cptr = WhoPtr; *cptr && sptr < zptr; *sptr++ = *cptr++);
      for (cptr = "\nmailto:"; *cptr && sptr < zptr; *sptr++ = *cptr++);
      for (cptr = WhoEmailPtr; *cptr && sptr < zptr; *sptr++ = *cptr++);
      if (sptr < zptr) *sptr++ = '\n';
      for (cptr = HostNamePtr; *cptr && sptr < zptr; *sptr++ = *cptr++);
      for (cptr = " ["; *cptr && sptr < zptr; *sptr++ = *cptr++);
      for (cptr = RemoteHostPtr; *cptr && sptr < zptr; *sptr++ = *cptr++);
      for (cptr = "]\n\n"; *cptr && sptr < zptr; *sptr++ = *cptr++);
      for (cptr = TextPtr; *cptr && sptr < zptr; *sptr++ = *cptr++);
      if (sptr >= zptr) ErrorExit (SS$_BUGCHECK, FI_LI);
      *sptr = '\0';

      cptr = SendMailMessage (MailFrom, CommentEmail, NULL,
                              SubjText, MailTextPtr);
      if (cptr)
      {
         TemplateAdvise ("E!AZ", cptr);
         /* put it into a cookie so it's available after the redirect */
         TemplateAdvise (NULL);
      }
   }

   /* try to prevent reloading POSTs which might duplicate actions */
   for (cptr = sptr = CgiRequestUriPtr; *sptr; sptr++);
   *sptr = '\0';
   CgiLibResponseRedirect ("%!%s\n%s", cptr, CookieHeader());

   return (TRUE);
}

/*****************************************************************************/
/*
Establish that the section does actually use $$COMMENT$$ and that the supplied
name is matched against it (or one of them if multiple).  Return NULL if not
found or a pointer to the start of the $$COMMENT if it is.
*/

char* CommentSubmitFind (char* FindCommentName)

{
   char  *cptr, *sptr,
         *CommentPtr = NULL;
   char  CommentName [256];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "CommentSubmitFind() |%s|\n", FindCommentName);

   cptr = ContentSectionPtr->TextPtr;
   while (*cptr)
   {
      if (!*(cptr = sptr = TemplateParseFind (cptr, "$$COMMENT$"))) break;
      /* if not named then return the first found */
      if (!CommentPtr) CommentPtr = cptr;
      cptr = TemplateParseFindParam (cptr+10, "name=");
      if (*(USHORTPTR)cptr != '$$')
         cptr = TemplateParseParam (cptr+5, CommentName, sizeof(CommentName));
      if (CommentName[0] == '~') return (sptr);
      if (strsame (CommentName, FindCommentName, -1)) return (sptr);
      if (*cptr) cptr+= 2;
   }

   return (CommentPtr);
}

/*****************************************************************************/
/*
Assess the submitted comment against some basic criteria (if configured).
1. Comment size in characters is not excessive.
2. Current comment file size is not excessive.
3. User-agent is present and does not contain a rejected string.
4. Comment who does not contain a link.
5. Comment text does not contain a 'reject' word or string.
6. A minimum interval has expired between comments from this host.
7. Total number of comments from this host is not excessive.
*/

BOOL CommentSubmitVet
(
char *FileNamePtr,
char *RemoteHostPtr,
char *CommentWhoPtr,
char *CommentTextPtr
)
{
   int  status,
        HostCommentCount = 0;
   unsigned long  LastComment;
   char  *cptr, *sptr, *zptr,
         *CommentPtr,
         *TextPtr;
   char  HostAddress [32],
         WhenString [64];
   stat_t  StatBuffer;
   struct tm  time_struct;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "CommentSubmitVet()\n");

   if (CommentRejectTextLength (CommentTextPtr,
                                CONFIG_LANG_PARSE(LangModerate,2)))
      return (FALSE);

   /*************/
   /* file size */
   /*************/

   if (stat (FileNamePtr, &StatBuffer) >= 0)
   {
      if (Debug) fprintf (stdout, "%d bytes\n", StatBuffer.st_size);
      if (StatBuffer.st_size >> 10 > VWcmsConfig.CommentMaxSizeKbytes)
      {
         if (WatchEnabled)
            WatchThis (FI_LI, "REJECT file !ULkB > !ULkB",
                       StatBuffer.st_size >> 10,
                       VWcmsConfig.CommentMaxSizeKbytes);
         TemplateAdvise ("W!AZ (!AZ %X!8XL)",
                         CONFIG_LANG_PARSE(LangModerate,2),
                         SysGetMsg(SS$_EXDISKQUOTA), SS$_EXDISKQUOTA);
         return (FALSE);
      }
   }
   else
   if (vaxc$errno != SS$_NOSUCHFILE && vaxc$errno != RMS$_FNF)
      ErrorExit (vaxc$errno, FI_LI);

   /**********/
   /* reject */
   /**********/

   if (CommentRejectWho (CommentWhoPtr, CONFIG_LANG_PARSE(LangModerate,2)))
      return (FALSE);

   if (CommentRejectString (CommentTextPtr, CONFIG_LANG_PARSE(LangModerate,2)))
      return (FALSE);

   if (CommentRejectWord (CommentTextPtr, CONFIG_LANG_PARSE(LangModerate,2)))
      return (FALSE);

   /*********************/
   /* existing comments */
   /*********************/

   /* only go to this trouble if configured! */
   if (!VWcmsConfig.CommentMinInterval &&
       !VWcmsConfig.CommentHostMax)
      return (TRUE);

   status = WaitForLock (FileNamePtr, FALSE);
   if (VMSnok (status))
   {
      TemplateAdvise ("ECOMMENT !AZ (%X!8XL)", SysGetMsg(status), status);
      return (FALSE);
   }

   status = DataReadFileIntoMemory (FileNamePtr, &CommentPtr, NULL, TRUE);

   WaitForUnLock ();

   if (VMSnok (status))
   {
      /* the there is no current file then we can't assess against it */
      if (status == RMS$_FNF) return (TRUE);

      TemplateAdvise ("W!AZ (!AZ %X!8XL)",
                      CONFIG_LANG_PARSE(LangModerate,2),
                      SysGetMsg(status), status);
      return (FALSE);
   }

   cptr = CommentPtr;
   while (*cptr)
   {
      /*****************/
      /* parse comment */
      /*****************/

      while (*cptr && *cptr != '<') cptr++;
      if (!*cptr) break;
      if (strncmp (cptr++, "<comment id=\"", 13)) continue;
      cptr += 12;
      while (*cptr && *cptr != '<') cptr++;

      if (strncmp (cptr, "<host>", 6)) continue;
      cptr += 6;
      while (*cptr && *cptr != '<' && !isspace(*cptr)) cptr++;
      while (*cptr && isspace(*cptr)) cptr++;
      zptr = (sptr = HostAddress) + sizeof(HostAddress)-1;
      while (*cptr && *cptr != '<' && sptr < zptr) *sptr++ = *cptr++;
      if (sptr >= zptr) continue;
      *sptr = '\0';

      /* if not from the same host */
      if (strcmp (HostAddress, RemoteHostPtr)) continue;

      HostCommentCount++;

      while (*cptr && *cptr != '>') cptr++;
      while (*cptr && *cptr != '<') cptr++;
      if (strncmp (cptr, "<when>", 6)) continue;
      cptr += 6;
      zptr = (sptr = WhenString) + sizeof(WhenString)-1;
      while (*cptr && *cptr != '<' && sptr < zptr) *sptr++ = *cptr++;
      if (sptr >= zptr) continue;
      *sptr = '\0';

      if (!GetTimeString (WhenString, &time_struct)) continue;
      LastComment = mktime (&time_struct);
      if ((int)LastComment == -1) continue;

      if (VWcmsConfig.CommentMinInterval &&
          CurrentSecondsTime - LastComment < VWcmsConfig.CommentMinInterval)
      {
         if (WatchEnabled)
            WatchThis (FI_LI, "REJECT interval <!ULS",
                       VWcmsConfig.CommentMinInterval);
         TemplateAdvise ("W!AZ (<!ULS)",
                          CONFIG_LANG_PARSE(LangModerate,2),
                          VWcmsConfig.CommentMinInterval);
         return (FALSE);
      }
   }

   if (VWcmsConfig.CommentHostMax &&
       HostCommentCount > VWcmsConfig.CommentHostMax)
   {
      if (WatchEnabled)
         WatchThis (FI_LI, "REJECT maximum host >!UL",
                    VWcmsConfig.CommentHostMax);
      TemplateAdvise ("W!AZ (>!UL)",
                       CONFIG_LANG_PARSE(LangModerate,2),
                       VWcmsConfig.CommentHostMax);
      return (FALSE);
   }

   return (TRUE);
}

/*****************************************************************************/
/*
Return true if the 'who' string is unacceptable.  False otherwise.
*/

BOOL CommentRejectWho
(
char *WhoPtr,
char *RejectMsg
)
{
   char  *cptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "CommentRejectWho()\n");

   if (strstr (WhoPtr, cptr = "http://") ||
       strstr (WhoPtr, cptr = "https://"))
   {
      /* characteristic SPAMbot */
      if (WatchEnabled) WatchThis (FI_LI, "REJECT who \'!AZ\'", cptr);
      TemplateAdvise ("W!AZ (\"!AZ\")", RejectMsg, WhoPtr);
      return (TRUE);
   }

   return (FALSE);
}

/*****************************************************************************/
/*
Return true if the text is an unacceptable length.  False otherwise.
*/

BOOL CommentRejectTextLength
(
char *TextPtr,
char *RejectMsg
)
{
   int  CommentTextLength;
   char  *cptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "CommentRejectTextLength()\n");

   CommentTextLength = strlen(TextPtr);

   /* surely anything worth it has got to be more than a few words! */
   if (CommentTextLength <= 16)
   {
      if (WatchEnabled) WatchThis (FI_LI, "REJECT text <16B");
      TemplateAdvise ("W!AZ (<16B)", RejectMsg);
      return (TRUE);
   }

   if (CommentTextLength > VWcmsConfig.CommentMaxChars)
   {
      if (WatchEnabled)
         WatchThis (FI_LI, "REJECT text >!ULB", VWcmsConfig.CommentMaxChars);
      TemplateAdvise ("W!AZ (>!ULB)", RejectMsg, VWcmsConfig.CommentMaxChars);
      return (TRUE);
   }

   return (FALSE);
}

/*****************************************************************************/
/*
Return true if one of the newline-delimitted [comment-reject-string] is found
in the supplied text.  False if not configured or not found.
*/

BOOL CommentRejectString
(
char *TextPtr,
char *RejectMsg
)
{
   char  *cptr, *sptr, *zptr;
   char  RejectString [256];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "CommentRejectString()\n");

   if (!(cptr = VWcmsConfig.CommentRejectString)) return (FALSE);

   while (*cptr)
   {
      while (*cptr && *cptr == '\n') cptr++;
      if (!*cptr) break;
      zptr = (sptr = RejectString) + sizeof(RejectString)-1;
      while (*cptr && *cptr != '\n' && sptr < zptr) *sptr++ = *cptr++;
      *sptr = '\0';
      while (*cptr && *cptr != '\n') cptr++;
      if (strstr (TextPtr, RejectString))
      {
         if (WatchEnabled)
            WatchThis (FI_LI, "REJECT string \'!AZ\'", RejectString);
         TemplateAdvise ("W!AZ (\"!AZ\")", RejectMsg, RejectString);
         return (TRUE);
      }
   }

   return (FALSE);
}

/*****************************************************************************/
/*
Return true if one of the space-delimitted [comment-reject-word] is found
in the supplied text.  False if not configured or not found.
*/

BOOL CommentRejectWord
(
char *TextPtr,
char *RejectMsg
)
{
   char  *cptr, *sptr, *zptr;
   char  RejectString [256];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "CommentRejectWord()\n");

   if (!VWcmsConfig.CommentRejectWord) return (FALSE);

   /* remove all but alphabetics, collapsing white-space, etc. */
   cptr = TextPtr;
   sptr = TextPtr = CgiLibVeeMemCalloc (strlen(TextPtr));
   if (!sptr) ErrorExit (vaxc$errno, FI_LI);
   while (*cptr)
   {
      while (*cptr && (ispunct(*cptr) || isspace(*cptr) || isdigit(*cptr)))
          cptr++;
      while (*cptr && !ispunct(*cptr) && !isspace(*cptr) && !isdigit(*cptr))
          *sptr++ = tolower(*cptr++);
   }
   *sptr = '\0';

   cptr = VWcmsConfig.CommentRejectWord;
   while (*cptr)
   {
      while (*cptr && isspace(*cptr)) cptr++;
      if (!*cptr) break;
      zptr = (sptr = RejectString) + sizeof(RejectString)-1;
      while (*cptr && !isspace(*cptr) && sptr < zptr)
         *sptr++ = tolower(*cptr++);
      *sptr = '\0';
      if (strstr (TextPtr, RejectString))
      {
         if (WatchEnabled)
            WatchThis (FI_LI, "REJECT word \'!AZ\'", RejectString);
         TemplateAdvise ("W!AZ (\"!AZ\")", RejectMsg, RejectString);
         return (TRUE);
      }
   }

   return (FALSE);
}

/*****************************************************************************/
/*
When the [moderate] button is pressed this function locks and loads (!) the
comemnts file, scans it comparing the unique ID in each item to checked boxes
(or rather the CGI variables that represent those) and modifies the moderation
character appropriately.  It then overwrites the file with the modified content
and unlocks the comment file.
*/

void CommentModerate ()

{
   int  status;
   char  *cptr, *sptr, *zptr,
         *CommentPtr,
         *CommentNamePtr,
         *ModerationPtr,
         *FileNamePtr;
   char  CommentId [16+1],
         FormName [64];
   FILE  *FilePtr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "CommentModerate()\n");

   if (ConfigReadOnly (TRUE)) return;

   if (!ManageSiteOn) return;

   CommentNamePtr = GetCgiVarNull ("FORM_COMMENT_NAME");
   if (!CommentNamePtr) ErrorExit (SS$_BUGCHECK, FI_LI);

   if (!(cptr = CommentSubmitFind (CommentNamePtr)))
   {
      /* couldn't find it ... */ 
      TemplateAdvise ("E$$COMMENT$!AZ$$??", CommentNamePtr);
      sleep (REJECT_SLEEP);
      return;
   }

   FileNamePtr = MakeFileName (CommentNamePtr, ".CMT");

   status = WaitForLock (FileNamePtr, TRUE);
   if (VMSnok (status))
   {
      TemplateAdvise ("ECOMMENT !AZ (%X!8XL)", SysGetMsg(status), status);
      return;
   }

   status = DataReadFileIntoMemory (FileNamePtr, &CommentPtr, NULL, TRUE);
   if (VMSnok (status))
   {
      WaitForUnLock ();
      if (status == RMS$_FNF) return;
      TemplateAdvise ("ECOMMENT !AZ (%X!8XL)", SysGetMsg(status), status);
      return;
   }

   cptr = CommentPtr;
   while (*cptr)
   {
      /*****************/
      /* parse comment */
      /*****************/

      while (*cptr && *cptr != '<') cptr++;
      if (!*cptr) break;
      if (strncmp (cptr++, "<comment id=\"", 13)) continue;
      cptr += 12;
      /* exactly 16 hexdecimal characters */
      zptr = (sptr = CommentId) + sizeof(CommentId)-1;
      while (*cptr && *cptr != '\"' && sptr < zptr) *sptr++ = *cptr++;
      if (*cptr != '\"') continue;
      *sptr = '\0';
      for (cptr++; *cptr && isspace(*cptr); cptr++);
      if (strncmp (cptr, "moderation=\"", 12)) continue;
      cptr += 12;
      ModerationPtr = cptr++;
      if (*cptr != '\"') continue;
      while (*cptr && *cptr != '>') cptr++;
      if (!*cptr) break;

      sprintf (FormName, "FORM_MOD_TICK_%s", CommentId);
      if (GetCgiVarNull (FormName))
         *ModerationPtr = '!';
      else
      {
         sprintf (FormName, "FORM_MOD_CUT_%s", CommentId);
         if (GetCgiVarNull (FormName))
            *ModerationPtr = 'X';
         else
         {
            sprintf (FormName, "FORM_MOD_UNCUT_%s", CommentId);
            if (GetCgiVarNull (FormName))
               *ModerationPtr = 'x';
         }
      }
   }

   FilePtr = fopen (FileNamePtr, "r+", "ctx=rec",
                    "rfm=stmlf", "shr=nil", "fop=cif");
   if (!FilePtr)
   {
      status = vaxc$errno;
      WaitForUnLock ();
      TemplateAdvise ("ECOMMENT !AZ (%X!8XL)", SysGetMsg(status), status);
      return;
   }

   fputs (CommentPtr, FilePtr);

   fclose (FilePtr);

   if (chmod (FileNamePtr, 0700)) ErrorExit (vaxc$errno, FI_LI);

   WaitForUnLock ();
}

/*****************************************************************************/

